{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Loop where {

import qualified ErrorMonad;
import qualified List;
import qualified Term;
import qualified HOL;
import qualified Nat;
import qualified Product_Type;
import qualified Trs;
import qualified Position;
import qualified Show;
import qualified Util;

rewrites ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => (Term.Term a b, Term.Term a b) ->
                         Term.Ctxt a b ->
                           Term.Sub a b ->
                             (Term.Term a b, Term.Term a b) ->
                               [(Term.Term a b, Term.Term a b)] ->
                                 Util.Eithera [Char] ();
rewrites (s, t) c sigma (l, r) ra =
  ErrorMonad.checkB
    (HOL.eqop s (Term.ctxt_apply c (Term.sub_apply l sigma)) &&
      (HOL.eqop t (Term.ctxt_apply c (Term.sub_apply r sigma)) &&
        List.member (l, r) ra))
    (List.append
      (List.append (List.append "Term " (Term.sshow_term s))
        " does not rewrite to term ")
      (Term.sshow_term t));

subst_safe ::
  forall a b.
    (Eq a,
      Eq b) => [a] ->
                 [(Term.Term b a, Term.Term b a)] ->
                   Term.Sub b a -> Maybe (Term.Sub b a);
subst_safe vs ((Term.Var x, t) : ts) sigma =
  (if List.member (Term.Var x, t) ts then Loop.subst_safe vs ts sigma
    else (if List.member x vs &&
               not (HOL.eqop (Term.sub_apply (Term.Var x) sigma) t)
           then Nothing
           else (case Loop.subst_safe (x : vs) ts sigma of {
                  Nothing -> Nothing;
                  Just a ->
                    let {
                      (Term.Sub s) = a;
                    } in Just (Term.Sub
                                (\ y -> (if HOL.eqop x y then t else s y)));
                })));
subst_safe vs [] sigma = Just sigma;

decomp :: forall a b. [a] -> [b] -> Maybe [(a, b)];
decomp (s : ss) (t : ts) =
  (case Loop.decomp ss ts of {
    Nothing -> Nothing;
    Just pairs -> Just ((s, t) : pairs);
  });
decomp [] (uw : ux) = Nothing;
decomp (uu : uv) [] = Nothing;
decomp [] [] = Just [];

decompose_one ::
  forall a b.
    (Eq a) => Term.Term a b ->
                Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose_one (Term.Fun f ts) (Term.Fun g ss) =
  (if HOL.eqop f g then Loop.decomp ts ss else Nothing);
decompose_one (Term.Fun v va) (Term.Var y) = Nothing;
decompose_one (Term.Var x) t = Just [(Term.Var x, t)];

decompose_n ::
  forall a b.
    (Eq a) => Nat.Nat ->
                Term.Term a b ->
                  Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose_n (Nat.Suc n) s t =
  (case Loop.decompose_n n s t of {
    Nothing -> Nothing;
    Just a ->
      List.foldla
        (\ aa (b @ (sa, ta)) ->
          (case aa of {
            Nothing -> Nothing;
            Just ys ->
              (case Loop.decompose_one sa ta of {
                Nothing -> Nothing;
                Just xs -> Just (List.append xs ys);
              });
          }))
        (Just []) a;
  });
decompose_n Nat.Zero_nat s t = Loop.decompose_one s t;

decompose ::
  forall a b.
    (Eq a) => Term.Term a b ->
                Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose s t = Loop.decompose_n (Term.sz s) s t;

decompose_list ::
  forall a b.
    (Eq a) => [(Term.Term a b, Term.Term a b)] ->
                Maybe [(Term.Term a b, Term.Term a b)];
decompose_list ts =
  List.foldla
    (\ a (b @ (s, t)) ->
      (case a of {
        Nothing -> Nothing;
        Just ls ->
          (case Loop.decompose s t of {
            Nothing -> Nothing;
            Just ys -> Just (List.append ls ys);
          });
      }))
    (Just []) ts;

match_all ::
  forall a b.
    (Eq a, Eq b) => [(Term.Term a b, Term.Term a b)] -> Maybe (Term.Sub a b);
match_all ts =
  (case Loop.decompose_list ts of {
    Nothing -> Nothing;
    Just ls -> Loop.subst_safe [] ls (Term.Sub Term.Var);
  });

getpr ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Term.Term a b ->
                         Term.Term a b ->
                           Position.Pos ->
                             (Term.Term a b, Term.Term a b) ->
                               Maybe (Term.Term a b,
                                       (Term.Ctxt a b,
 (Term.Sub a b, (Term.Term a b, Term.Term a b))));
getpr s t p (l, r) =
  (case Loop.match_all [(l, Term.subt_at s p), (r, Term.subt_at t p)] of {
    Nothing -> Nothing;
    Just sigma ->
      (if ErrorMonad.isValue
            (Loop.rewrites (s, t) (Trs.ctxt_of_pos_term p s) sigma (l, r)
              [(l, r)])
        then Just (s, (Trs.ctxt_of_pos_term p s, (sigma, (l, r))))
        else Nothing);
  });

getp ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Term.Term a b ->
                         Term.Term a b ->
                           Position.Pos ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Maybe (Term.Term a b,
                                       (Term.Ctxt a b,
 (Term.Sub a b, (Term.Term a b, Term.Term a b))));
getp s t p (r : rs) =
  (case Loop.getpr s t p r of {
    Nothing -> Loop.getp s t p rs;
    Just a -> Just a;
  });
getp s t p [] = Nothing;

find_get_p ::
  forall f v.
    (Eq f, Show.Sshow f, Eq v,
      Show.Sshow v) => Term.Term f v ->
                         Term.Term f v ->
                           [Position.Pos] ->
                             [(Term.Term f v, Term.Term f v)] ->
                               Maybe (Term.Term f v,
                                       (Term.Ctxt f v,
 (Term.Sub f v, (Term.Term f v, Term.Term f v))));
find_get_p s t (p : ps) r =
  (case Loop.getp s t p r of {
    Nothing -> Loop.find_get_p s t ps r;
    Just a -> Just a;
  });
find_get_p s t [] r = Nothing;

get ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Term.Term a b ->
                         Term.Term a b ->
                           [(Term.Term a b, Term.Term a b)] ->
                             Util.Eithera [Char]
                               (Term.Term a b,
                                 (Term.Ctxt a b,
                                   (Term.Sub a b,
                                     (Term.Term a b, Term.Term a b))));
get s t r =
  (case Loop.find_get_p s t (Term.poss_list s) r of {
    Nothing ->
      ErrorMonad.errora
        (List.append
          (List.append
            (List.append
              (List.append "Could not find reduction between term "
                (Term.sshow_term s))
              " and term ")
            (Term.sshow_term t))
          ".");
    Just a -> ErrorMonad.returna a;
  });

get_list ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b)] ->
                         Term.Term a b ->
                           [(Term.Term a b, Term.Term a b)] ->
                             Util.Eithera [Char]
                               [(Term.Term a b,
                                  (Term.Ctxt a b,
                                    (Term.Sub a b,
                                      (Term.Term a b, Term.Term a b))))];
get_list (s : t : ts) u r =
  ErrorMonad.bind (Loop.get s t r)
    (\ d ->
      ErrorMonad.bind (Loop.get_list (t : ts) u r)
        (\ y -> ErrorMonad.returna (d : y)));
get_list [s] u r =
  ErrorMonad.bind (Loop.get s u r) (\ y -> ErrorMonad.returna [y]);
get_list [] u r = ErrorMonad.errora "Loop must be non-empty";

get_loop ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b)] ->
                         Term.Ctxt a b ->
                           Term.Sub a b ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Util.Eithera [Char]
                                 [(Term.Term a b,
                                    (Term.Ctxt a b,
                                      (Term.Sub a b,
(Term.Term a b, Term.Term a b))))];
get_loop (t : ts) d tau r =
  Loop.get_list (t : ts) (Term.ctxt_apply d (Term.sub_apply t tau)) r;
get_loop [] uu uv uw = ErrorMonad.errora "Loop must be non-empty";

rewrites_to ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b,
                          (Term.Ctxt a b,
                            (Term.Sub a b, (Term.Term a b, Term.Term a b))))]
                         -> Term.Term a b ->
                              [(Term.Term a b, Term.Term a b)] ->
                                Util.Eithera [Char] ();
rewrites_to ((s, (c, (sigma, rule))) : (t, (d, (tau, rule'))) : xs) u r =
  ErrorMonad.bind (Loop.rewrites (s, t) c sigma rule r)
    (\ x -> Loop.rewrites_to ((t, (d, (tau, rule'))) : xs) u r);
rewrites_to [(s, (c, (sigma, rule)))] t r = Loop.rewrites (s, t) c sigma rule r;

check_loop_d ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b,
                          (Term.Ctxt a b,
                            (Term.Sub a b, (Term.Term a b, Term.Term a b))))]
                         -> Term.Ctxt a b ->
                              Term.Sub a b ->
                                [(Term.Term a b, Term.Term a b)] ->
                                  Util.Eithera [Char] ();
check_loop_d (v : va) d tau r =
  Loop.rewrites_to (v : va)
    (Term.ctxt_apply d (Term.sub_apply (fst (List.hd (v : va))) tau)) r;
check_loop_d [] uu uv uw =
  ErrorMonad.errora "There must be at least one rewrite step";

check_loop ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => ([(Term.Term a b)], (Term.Ctxt a b, Term.Sub a b)) ->
                         [(Term.Term a b, Term.Term a b)] ->
                           Util.Eithera [Char] ();
check_loop (ts, (c, sigma)) r =
  ErrorMonad.bind (Loop.get_loop ts c sigma r)
    (\ ls -> Loop.check_loop_d ls c sigma r);

}
